/*
 * BinarySearchTree.java
    * This class illustrates optimal BSTs as developed in most
    * texts.  See for example, Neopolitan or Baase.
    * It includes the possibility that some search keys might not
    * be in the tree.
 * @author  Stuart Hansen
 * @version May 8, 2001
 */

import java.io.*;
import java.util.*;
import Nodes.*;
import Keys.*;

public class OptimalBSTwithMisses {

    INode root;           // The root of the tree
    ArrayList key;        // The array list of keys
    ArrayList prob;       // The array list of probabilities
    
    double ASTPrime = 0.0;    // initializing the ASTPrime value
    
    /** Creates new BinarySearchTree */
    public OptimalBSTwithMisses() {
        root = NullNode.getInstance();
    }
  
    /** Build an optimal binary search tree from a file of values
     *  and probabilities */
    public void buildOptimalFromFile (String filename) { 
        String line = "";
        String probstr = "";
        key = new ArrayList();
        key.add(null);
        prob = new ArrayList();
        prob.add(null);
        try {

            File inFile = new File (filename);
            FileReader fileReader = new FileReader(inFile);
            BufferedReader bufReader = new BufferedReader(fileReader);
        
            /** Read in a value and a probability and insert them into array lists */
            line = bufReader.readLine(); 
            while (line != null) {                
                int value = Integer.parseInt(line);
                key.add(new Integer(value));
                probstr = bufReader.readLine();
                double probab = Double.parseDouble(probstr);
                prob.add(new Double(probab));
                line = bufReader.readLine();
            }
        }
        catch (EOFException e) {
            System.out.println("At End of file.");
        }
        catch (IOException e) {
            System.out.println(e.toString());
        }
        catch (NumberFormatException e){
            System.out.println ("Unable to parse " + line);
        }
        
        /** Use the array lists to create an optimal binary search tree
         *   This code is borrowed from Neapolitan, beginning on page 120 */
        int n = key.size()-1;
        double A [][] = new double [n+2][n+1];
        double probSum[][]= new double[n+2][n+1];
        int R [][] = new int [n+2][n+1];
    

        int i, j, k;  // counters
        double t;
                
        for (i=1; i<=n; i++){
            A[i][i] = ((Double)prob.get(i)).doubleValue();
            A[i][i-1] = 0;
            R[i][i] = i;
            R[i][i-1] = 0;
        }
        
        A[n+1][n]=0;
        R[n+1][n]=0;
        
        for (int diag=1; diag<=n; diag++)
            for (i=1; i<=n-diag; i++) {
                j = i+diag;
                double sum = 0.0;
                
                // This should be made much more efficient,
                // but it works.
                // Find the minimum cost for location [i][j]
                for (k=i; k<=j; k++) 
                    sum = sum + ((Double)prob.get(k)).doubleValue();
                A[i][j] = A[i][i-1]+A[i+1][j]+ sum;
                R[i][j] = i;
                for (k=i+1; k<=j; k++){
                    if (A[i][k-1] + A[k+1][j] + sum < A[i][j]) {
                        A[i][j] = A[i][k-1] + A[k+1][j] + sum;
                        R[i][j] = k;
                    }
                }
            }
        
        root = tree(1, n, A, R, probSum);
    }                   
    
    /** Build an optimal binary search tree from a file of values
     *  and probabilities, including probabilities of misses */
    public void buildOptimalWithMissesFromFile (String filename) { 
        String line = "";
        String probstr = "";
        key = new ArrayList();
        key.add(null);
        prob = new ArrayList();
        prob.add(null);
        ArrayList  q = new ArrayList();
        try {

            File inFile = new File (filename);
            FileReader fileReader = new FileReader(inFile);
            BufferedReader bufReader = new BufferedReader(fileReader);
        
            /** Read in a value and a probability and insert them into array lists */
            probstr = bufReader.readLine(); 
            double probab2 = Double.parseDouble(probstr);
            q.add(new Double(probab2));     
            line = bufReader.readLine();
            while (line != null) {  
                int value = Integer.parseInt(line);
                key.add(new Integer(value));
                
                probstr = bufReader.readLine();
                double probab = Double.parseDouble(probstr);
                prob.add(new Double(probab));
                
                probstr = bufReader.readLine();
                probab2 = Double.parseDouble(probstr);
                q.add(new Double(probab2));
                
                line = bufReader.readLine();
            }
        }
        catch (EOFException e) {
            System.out.println("At End of file.");
        }
        catch (IOException e) {
            System.out.println(e.toString());
        }
        catch (NumberFormatException e){
            System.out.println ("Unable to parse " + line);
        }
        
        /** Use the array lists to create an optimal binary search tree
         *   This code is borrowed from Neapolitan, beginning on page 120 */
        int n = key.size()-1;
        double A [][] = new double [n+2][n+1];
        double probSum[][] = new double [n+2][n+1];
        int R [][] = new int [n+2][n+1];
    

        int i, j, k;  // counters
        double t;
                
        for (i=1; i<=n+1; i++){
            A[i][i-1] = ((Double)q.get(i-1)).doubleValue();
            R[i][i-1] = 0;
        }
        
        for(int l=1; l<=n+1; l++)
            for (int m=0; m<=n; m++) {
                double probS = 0.0;
                for (int p=l; p<=m; p++)
                    probS += ((Double)prob.get(p)).doubleValue();
                for (int p=l-1; p<=m; p++)
                    probS += ((Double)q.get(p)).doubleValue();
                probSum[l][m] = probS;
            }
                 
        for (int diag=0; diag<=n; diag++)
            for (i=1; i<=n-diag; i++) {
                j = i+diag;
                double sum = 0.0;
                
                // This should be made much more efficient,
                // but it works.
                // Find the minimum cost for location [i][j]
                for (k=i; k<=j; k++) 
                    sum = sum + ((Double)prob.get(k)).doubleValue();
                
                double qsum = 0.0;
                for (k=i-1; k<=j; k++)
                    qsum = qsum + ((Double)q.get(k)).doubleValue();
                
                A[i][j] = A[i][i-1]+A[i+1][j] + sum + qsum;
                R[i][j] = i;
                for (k=i+1; k<=j; k++){
                    if (A[i][k-1] + A[k+1][j] + sum + qsum < A[i][j]) {
                        A[i][j] = A[i][k-1] + A[k+1][j] + sum + qsum;
                        R[i][j] = k;
                    }
                }
            }
        System.out.println ("The Average Search Time = " + A[1][n]);  
        root = tree(1, n, A, R, probSum);
        System.out.println ("ASTPrime = " + ASTPrime);
    }  
    

    
    /** Actually build the tree from the A and R arrays.
     *  this code seems very UNobject-oriented, as we are
     *  setting left and right references someplace besides
     *  in the the Node */
    INode tree(int i, int j, double A[][], int R[][], double probSum[][]) {
        int k;
        
        k = R[i][j];
        if (k==0)
            return (NullNode.getInstance());
        Node p = new Node (new IntegerKey ((Integer)key.get(k)));
        ASTPrime += probSum[i][k-1] + 2.0*(probSum[k+1][j] + ((Double)prob.get(k)).doubleValue());
        p.setLeft(tree(i, k-1, A, R, probSum));
        p.setRight(tree(k+1, j, A, R, probSum));
        return p;
    }
    
    /** search the tree starting at the root */
    public boolean search (IntegerKey search_key) {
        return root.search(search_key);
    }
    
    /** Traverse the tree printing out each element */
    public void inOrder() {
        root.inOrder();
    }
    
    /** Traverse the tree in preOrder printing out as we go */
    public void preOrder() {
        root.preOrder();
    }


    /** a test main for this class (and its subclasses) */
    public static void main (String args[])
    {
        OptimalBSTwithMisses bst = new OptimalBSTwithMisses();
        bst.buildOptimalWithMissesFromFile("/home/hansen/Research/optimalBST/data/TestExtension4.txt");
        bst.preOrder();
    }                
}







