/*
 * BinarySearchTreeWithRange.java
    * This file implements optimal BSTs with range nodes instead of NullNodes.
 *
 * Created on May 8, 2001, 11:00 AM
 * @author Stuart Hansen
 */

import java.io.*;
import java.util.*;
import Keys.*;
import Nodes.*;

/**
 *
 * @author  Stuart Hansen
 * @version May 8, 2001
 */
public class OptimalBSTwithRange {

    INode root;  // The root of the tree
    ArrayList key;
    ArrayList prob;
    
    double ASTPrime = 0.0;
    
    /** Creates new BinarySearchTree */
    public OptimalBSTwithRange() {
        root = InfiniteRangeNode.getInstance();
    }
    
    /** Build an optimal binary search tree from a file of values
     *  and probabilities, including probabilities of misses */
    public void buildOptimalWithMissesFromFile (String filename) { 
        String line = "";
        String probstr = "";
        key = new ArrayList();
        key.add(null);
        prob = new ArrayList();
        prob.add(null);
        
        try {

            File inFile = new File (filename);
            FileReader fileReader = new FileReader(inFile);
            BufferedReader bufReader = new BufferedReader(fileReader);
        
            /** Read in a value and a probability and insert them into array lists */
            probstr = bufReader.readLine(); 
            double probab2 = Double.parseDouble(probstr);
            prob.add(new Double(probab2));     
            line = bufReader.readLine();
            while (line != null) {  
                int value = Integer.parseInt(line);
                key.add(new Integer(value));
                
                probstr = bufReader.readLine();
                double probab = Double.parseDouble(probstr);
                prob.add(new Double(probab));
                
                probstr = bufReader.readLine();
                probab2 = Double.parseDouble(probstr);
                prob.add(new Double(probab2));
                
                line = bufReader.readLine();
            }
        }
        catch (EOFException e) {
            System.out.println("At End of file.");
        }
        catch (IOException e) {
            System.out.println(e.toString());
        }
        catch (NumberFormatException e){
            System.out.println ("Unable to parse " + line);
        }
        
        /** Use the array lists to create an optimal binary search tree
         *   This code is borrowed from Neapolitan, beginning on page 120 */
        int n = prob.size()-1;
        double A [][] = new double [n+2][n+1];
        double probSum [][] = new double [n+2][n+1];
        int R [][] = new int [n+2][n+1];
    

        int i, j, k;  // counters
        double t;
                
        for (i=1; i<=n; i++){
            A[i][i] = ((Double)prob.get(i)).doubleValue();
            A[i][i-1] = 0;
            R[i][i] = i;
            R[i][i-1] = 0;
        }
        
        A[n+1][n]=0;
        R[n+1][n]=0;
        
        for (int l=1; l<=n; l++)
            for (int m=1; m<=n; m++) {
                double probS = 0.0;
                for (int p=l; p<=m; p++) 
                    probS += ((Double)prob.get(p)).doubleValue();
                probSum[l][m] = probS;
            }

        
        for (int diag=1; diag<=n; diag++)
            for (i=1; i<=n-diag; i++) {
                j = i+diag;
                double sum = 0.0;
                
                // This should be made much more efficient,
                // but it works.
                // Find the minimum cost for location [i][j]
                for (k=i; k<=j; k++) 
                    sum = sum + ((Double)prob.get(k)).doubleValue();
                
                
                A[i][j] = A[i][i-1]+A[i+1][j]+ sum;
                R[i][j] = i;
                for (k=i+1; k<=j; k++){
                    if (A[i][k-1] + A[k+1][j] + sum < A[i][j]) {
                        A[i][j] = A[i][k-1] + A[k+1][j] + sum;
                        R[i][j] = k;
                    }
                }
            }
        System.out.println ("The Average Search Time = " + A[1][n]);
        root = tree(1, n, A, R, probSum);
        System.out.println ("The AST Prime = " + ASTPrime);
    } 
    
    /** Actually build the tree from the A and R arrays.
     *  this code seems very UNobject-oriented, as we are
     *  setting left and right references someplace besides
     *  in the the Node */
    INode tree(int i, int j, double A[][], int R[][], double probSum[][]) {
        int k;
        INode p;
       
        k = R[i][j];

        if (k==0)
            return (NullNode.getInstance());

        double leftProbSum = probSum[i][k-1];
        double rightProbSum = probSum[k+1][j];
        double keyProb = ((Double)prob.get(k)).doubleValue();

        if (k == 1) {
            p = new NegInfRangeNode (new IntegerKey((Integer)key.get(k)));
            ASTPrime += ((Double)prob.get(k)).doubleValue();
            p.setRight(tree(k+1, j, A, R, probSum));
                    }
        else if (k == prob.size()-1) {
            p = new PosInfRangeNode (new IntegerKey((Integer)key.get(k/2)));
            ASTPrime += ((Double)prob.get(k)).doubleValue();
            p.setLeft(tree(i, k-1, A, R, probSum));           
        }
        else if (k%2 == 0) {
            p = new Node (new IntegerKey((Integer)key.get(k/2)));
            ASTPrime += leftProbSum + 2.0*(rightProbSum + keyProb);
            p.setLeft(tree(i, k-1, A, R, probSum));
            p.setRight(tree(k+1, j, A, R, probSum));
        }
        else {
            p = new RangeNode (new IntegerKey((Integer)key.get(k/2)), 
                               new IntegerKey((Integer)key.get((k+2)/2)));
            ASTPrime += leftProbSum + 2.0*(rightProbSum + keyProb);
            p.setLeft(tree(i, k-1, A, R, probSum));
            p.setRight(tree(k+1, j, A, R, probSum));
        }
        return p;
    }
    
    /** search the tree starting at the root */
    public boolean search (Key search_key) {
        return root.search(search_key);
    }
    
    /** Traverse the tree printing out each element */
    public void inOrder() {
        root.inOrder();
    }
    
    /** Traverse the tree in preOrder printing out as we go */
    public void preOrder() {
        root.preOrder();
    }


    /** a test main for this class (and its subclasses) */
    public static void main (String args[])
    {
       OptimalBSTwithRange bst = new OptimalBSTwithRange();
       bst.buildOptimalWithMissesFromFile("/home/hansen/Research/optimalBST/data/TestExtension4.txt");    
       bst.preOrder();
       System.out.println();
    }                
}







